import tensorflow as tf
from tensorflow.keras import layers

import wacky_rl
from wacky_rl import losses

class WackyModel(tf.keras.Model):

    def __init__(
            self,
            model_layer: (list, tf.keras.layers.Layer) = None,
            optimizer: tf.keras.optimizers.Optimizer = None,
            loss: losses.WackyLoss = None,
            loss_alpha: float = 1.0,
            model_name: str = 'UnnamedWackyModel',
            model_index: int = None,
            grad_clip=False,
    ):

        super().__init__()

        self.model_name = model_name
        self.model_index = model_index
        self.grad_clip = grad_clip

        # Loss Function:
        self.loss_alpha = loss_alpha

        if loss is None:
            self._wacky_loss = wacky_rl.losses.MeanSquaredErrorLoss()
        else:
            self._wacky_loss = loss
            # TODO: Add keras loss wrapper and str

        # Model Layer:
        if model_layer is None:
            self._wacky_layer = [
                layers.Flatten(),
                layers.Dense(64, activation='relu'),
                layers.Dense(64, activation='relu')
            ]
            # TODO: Add action layer based on loss

        else:
            if not isinstance(model_layer, list):
                self._wacky_layer = [model_layer]
            else:
                self._wacky_layer = model_layer

        # Optimizer:
        if optimizer is None:
            self._optimizer = tf.keras.optimizers.RMSprop()
        else:
            self._optimizer = optimizer
            # TODO: Add keras optimizer with str

    def _wacky_forward(self, x):
        for l in self._wacky_layer: x = l(x)
        return x

    def call(self, inputs, training=False, mask=None, *args, **kwargs):
        return self._wacky_forward(inputs)

    def train_step(self, inputs, *args, **kwargs):

        with tf.GradientTape() as tape:
            x = self._wacky_forward(inputs)
            loss = self.loss_alpha * self._wacky_loss(x, *args, **kwargs)

        grads = tape.gradient(loss, self.trainable_variables)

        if self.grad_clip:
            grads = [tf.clip_by_norm(g, 0.5) for g in grads]

        self._optimizer.apply_gradients(zip(grads, self.trainable_variables))

        return loss

    def predict_step(self, data, mask=None, *args, **kwargs):
        return self.call(data, training=False, mask=mask, *args, **kwargs)


class WackyDualingModel:

    def __init__(
            self,
            model_layer: (list, tf.keras.layers.Layer) = None,
            optimizer: tf.keras.optimizers.Optimizer = None,
            loss: wacky_rl.losses.WackyLoss = None,
            loss_alpha: float = 1.0,
            model_name: str = 'UnnamedWackyModel',
            model_index: int = None,
            grad_clip=False,
    ):

        self.model_1 = WackyModel(model_layer, optimizer, loss, loss_alpha, model_name+'_1', model_index, grad_clip)
        self.model_2 = WackyModel(model_layer, optimizer, loss, loss_alpha, model_name+'_2', model_index, grad_clip)

    @property
    def is_dualing(self):
        return True

    def __call__(self, inputs, training=True, mask=None, *args, **kwargs):
        x_1 = self.model_1(inputs, training, mask)
        x_2 = self.model_2(inputs, training, mask)
        return tf.math.minimum(x_1, x_2)

    def train_step(self, *args, **kwargs):
        loss_1 = self.model_1.train_step(*args, **kwargs)
        loss_2 = self.model_2.train_step(*args, **kwargs)
        return tf.reduce_mean([loss_1, loss_2], 0)

    def predict_step(self, data, mask=None, *args, **kwargs):
        return self(data, training=False, mask=mask, *args, **kwargs)


class TargetUpdate:

    def __init__(self, tau: float = 0.15):
        self.tau = tau

    def __call__(self, model: tf.keras.Model, target: tf.keras.Model):

        if hasattr(model, 'is_dualing'):
            if model.is_dualing:
                target.model_1 = self._update_target(target.model_1, model.model_1)
                target.model_2 = self._update_target(target.model_2, model.model_2)
                return target
        else:
            return self._update_target(model, target)

    def _update_target(self, model, target):
        weights = model.get_weights()
        target_weights = target.get_weights()

        for i in range(len(target_weights)):
            target_weights[i] = weights[i] * self.tau + target_weights[i] * (1 - self.tau)

        target.set_weights(target_weights)
        return target


class TargetModelWrapper:

    def __init__(self, model, tau: float = 0.15):

        import copy

        self.model = model
        self.target = copy.deepcopy(model)
        self._update_target = TargetUpdate(tau)

    def __call__(self, *args, **kwargs):
        return self.model( *args, **kwargs)

    def train_step(self, *args, **kwargs):
        return self.model.train_step(*args,**kwargs)

    def update_target(self):
        self.target_model = self._update_target(self.model, self.target)



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import tensorflow as tf

from tensorflow.keras.layers import Dense, Input
from tensorflow.keras import Model

from wacky_rl.agents import AgentCore
from wacky_rl.models import WackyModel
from wacky_rl.memory import BufferMemory
from wacky_rl.layers import DiscreteActionLayer, ContinActionLayer
from wacky_rl.losses import ActorLoss
from wacky_rl.transform import ExpectedReturnsCalculator
from wacky_rl.trainer import Trainer


class A2C(AgentCore):

    def __init__(self, env, approximate_contin=False):
        super(A2C, self).__init__()

        self.approximate_contin = approximate_contin
        self.memory = BufferMemory()
        self.calc_returns = ExpectedReturnsCalculator()

        # Actor:
        model_layer = [
            Dense(256, activation='relu'),
            Dense(256, activation='relu'),
        ]

        num_actions = int(self.decode_space(env.action_space))

        if self.space_is_discrete(env.action_space):
            model_layer.append(DiscreteActionLayer(num_bins=num_actions))
        elif self.approximate_contin:
            model_layer.append(DiscreteActionLayer(num_bins=21, num_actions=num_actions))
        else:
            model_layer.append(ContinActionLayer(num_actions=num_actions))

        self.actor = WackyModel(
            model_layer=model_layer,
            optimizer=tf.keras.optimizers.RMSprop(3e-4),
            loss=ActorLoss(entropy_factor=0.001),
            grad_clip=True,
        )

        # Critic:
        critic_input = Input(shape=env.observation_space.shape)
        critic_dense = Dense(256, activation='relu')(critic_input)
        critic_dense = Dense(256, activation='relu')(critic_dense)
        critic_out = Dense(1)(critic_dense)
        self.critic = Model(critic_input, critic_out)
        self.critic.compile(optimizer='adam', loss='mse')

    def act(self, inputs, act_argmax=False, save_memories=True):

        inputs = tf.expand_dims(tf.squeeze(inputs), 0)
        dist = self.actor(inputs, act_argmax=True)

        if act_argmax:
            actions = dist.mean_actions()
        else:
            actions = dist.sample_actions()

        if save_memories:
            self.memory(actions, key='actions')

        if self.approximate_contin:
            return dist.discrete_to_contin(actions).numpy()
        return actions.numpy()

    def learn(self):

        actions, states, new_states, rewards, dones = self.memory.replay()

        values = self.critic.predict(states)
        returns = self.calc_returns(rewards)

        adv = returns - values
        adv = tf.squeeze(adv)
        adv = (adv - tf.reduce_mean(adv)) / (tf.math.reduce_std(adv) + 1e-8)

        c_loss = self.critic.train_on_batch(states, returns)
        a_loss = self.actor.train_step(states, actions=actions, advantage=adv)

        self.memory.clear()

        return tf.reduce_mean(a_loss).numpy(), tf.reduce_mean(c_loss).numpy()


def train_a2c():

    import gym
    env = gym.make('CartPole-v0')
    # env = gym.make("LunarLanderContinuous-v2")
    agent = A2C(env)

    trainer = Trainer(env, agent)
    trainer.episode_train(300)
    trainer.test(100)

if __name__ == "__main__":
    train_a2c()
